from StringIO import StringIO
from colander import null
from pyramid.httpexceptions import HTTPMovedPermanently

from mock import MagicMock
import pytest

from kotti.testing import DummyRequest
from kotti.views.file import inline_view
from kotti.views.file import attachment_view
from kotti.views.file import UploadedFileResponse


class TestFileViews:
    def _create_file(self, config):
        from kotti.resources import File
        self.file = File("file contents", u"myf\xfcle.png", u"image/png")

    def _test_common_headers(self, headers):
        for name in ('Content-Disposition', 'Content-Length', 'Content-Type'):
            assert isinstance(headers[name], str)
        assert headers["Content-Length"] == "13"
        assert headers["Content-Type"] == "image/png"

    @pytest.mark.parametrize("params",
                             [(inline_view, 'inline'),
                              (attachment_view, 'attachment')])
    def test_file_views(self, params, config, filedepot):
        view, disposition = params
        self._create_file(config)
        res = view(self.file, None)

        self._test_common_headers(res.headers)

        assert res.headers["Content-Disposition"] == disposition + \
            ';filename="myfle.png"'

        assert res.app_iter.file.read() == 'file contents'
        res.app_iter.file.seek(0)
        assert res.body == 'file contents'


class TestFileEditForm:
    def make_one(self):
        from kotti.views.edit.content import FileEditForm
        from kotti.resources import File

        return FileEditForm(File(), DummyRequest())

    def test_edit_with_file(self, db_session, filedepot):
        view = self.make_one()
        view.edit(
            title=u'A title', description=u'A description',
            tags=[u"A tag"],
            file=dict(
                fp=StringIO('filecontents'),
                filename=u'myfile.png',
                mimetype=u'image/png',
                size=10,
                uid="randomabc",
                ),
            )
        assert view.context.title == u'A title'
        assert view.context.description == u'A description'
        assert view.context.data.file.read() == 'filecontents'
        assert view.context.filename == u'myfile.png'
        assert view.context.mimetype == u'image/png'
        assert view.context.size == len('filecontents')
        assert view.context.tags == [u"A tag"]

    def test_edit_without_file(self, filedepot):
        view = self.make_one()
        view.context.data = 'filecontents'
        view.context.filename = u'myfile.png'
        view.context.mimetype = u'image/png'
        view.context.size = 777
        view.edit(title=u'A title',
                  description=u'A description',
                  tags=[],
                  file=null)
        assert view.context.title == u'A title'
        assert view.context.description == u'A description'
        assert view.context.data.file.read() == 'filecontents'
        assert view.context.filename == u'myfile.png'
        assert view.context.mimetype == u'image/png'
        assert view.context.size == 777


class TestFileAddForm:
    def make_one(self):
        from kotti.views.edit.content import FileAddForm
        return FileAddForm(MagicMock(), DummyRequest())

    def test_add(self, config, filedepot):
        view = self.make_one()
        file = view.add(
            title=u'A title',
            description=u'A description',
            tags=[],
            file=dict(
                fp=StringIO('filecontents'),
                filename=u'myfile.png',
                mimetype=u'image/png',
                size=None,
                uid="randomabc",
                ),
            )

        assert file.title == u'A title'
        assert file.description == u'A description'
        assert file.tags == []
        assert file.data.file.read() == 'filecontents'
        assert file.filename == u'myfile.png'
        assert file.mimetype == u'image/png'
        assert file.size == len('filecontents')


class TestFileUploadTempStore:
    def make_one(self):
        from kotti.views.form import FileUploadTempStore
        return FileUploadTempStore(DummyRequest())

    def test_keys(self):
        tmpstore = self.make_one()
        tmpstore.session['important'] = 3
        tmpstore.session['_secret'] = 4
        assert tmpstore.keys() == ['important']

    def test_delitem(self):
        tmpstore = self.make_one()
        tmpstore.session['important'] = 3
        del tmpstore['important']
        assert 'important' not in tmpstore.session


class TestDepotStore:

    from kotti.resources import File, Image

    @pytest.mark.parametrize("factory", [File, Image])
    def test_create(self, factory, filedepot):
        f = factory('file content')
        assert len(f.data['files']) == 1
        assert f.data.file.read() == 'file content'

    @pytest.mark.parametrize("factory", [File, Image])
    def test_edit_content(self, factory, filedepot):
        f = factory('file content')
        assert f.data.file.read() == 'file content'
        f.data = 'edited'
        assert f.data.file.read() == 'edited'

    @pytest.mark.parametrize("factory", [File, Image])
    def test_session_rollback(self, factory, db_session, filedepot):
        from depot.manager import DepotManager

        f = factory(data='file content', name=u'content', title=u'content')
        id = f.data['file_id']

        db_session.add(f)
        db_session.flush()
        assert id in DepotManager.get()._storage.keys()

        db_session.rollback()
        assert id not in DepotManager.get()._storage.keys()
        assert DepotManager.get().delete.called

    @pytest.mark.parametrize("factory", [File, Image])
    def test_delete(self, factory, db_session, root, filedepot):
        from depot.manager import DepotManager

        f = factory(data='file content', name=u'content', title=u'content')
        id = f.data['file_id']
        root[str(id)] = f
        db_session.flush()

        assert id in DepotManager.get()._storage.keys()

        del root[str(id)]
        import transaction
        transaction.commit()

        assert DepotManager.get().delete.called
        assert id not in DepotManager.get()._storage.keys()


class TestUploadedFileResponse:
    def _create_file(self,
                     data="file contents",
                     filename=u"myf\xfcle.png",
                     mimetype=u"image/png"):
        from kotti.resources import File
        return File(data, filename, mimetype)

    def test_as_body(self, filedepot):
        f = self._create_file()
        resp = UploadedFileResponse(f.data, DummyRequest())
        assert resp.body == 'file contents'

    def test_as_app_iter(self, filedepot):
        from pyramid.response import FileIter

        f = self._create_file()
        resp = UploadedFileResponse(f.data, DummyRequest())
        assert isinstance(resp.app_iter, FileIter)
        assert ''.join(resp.app_iter) == 'file contents'

    def test_unknown_filename(self, filedepot):
        f = self._create_file("file contents", u"file", None)
        resp = UploadedFileResponse(f.data, DummyRequest())
        assert resp.headers['Content-Type'] == 'application/octet-stream'

    def test_guess_content_type(self, filedepot):
        f = self._create_file("file contents", u"file.jpg", None)
        resp = UploadedFileResponse(f.data, DummyRequest())
        assert resp.headers['Content-Type'] == 'image/jpeg'

    def test_caching(self, filedepot, monkeypatch):
        import datetime
        import webob.response

        f = self._create_file()
        d = datetime.datetime(2012, 12, 31, 13, 0, 0)

        class mockdatetime:
            @staticmethod
            def utcnow():
                return d

        monkeypatch.setattr(webob.response, 'datetime', mockdatetime)

        resp = UploadedFileResponse(f.data, DummyRequest(), cache_max_age=10)

        # this is set by filedepot fixture
        assert resp.headers['Last-Modified'] == 'Sun, 30 Dec 2012 00:00:00 GMT'
        assert resp.headers['Expires'] == 'Mon, 31 Dec 2012 13:00:10 GMT'

    def test_redirect(self, filedepot):
        f = self._create_file()
        f.data._thaw()
        f.data['_public_url'] = 'http://example.com'
        f.data._freeze()

        with pytest.raises(HTTPMovedPermanently) as e:
            UploadedFileResponse(f.data, DummyRequest())

        response = e.value
        assert response.headers['Location'] == 'http://example.com'



# -*- coding: utf-8 -*-

from pyramid.httpexceptions import HTTPMovedPermanently
from pyramid.response import _BLOCK_SIZE
from pyramid.response import FileIter
from pyramid.response import Response
from pyramid.view import view_config

from kotti.resources import File

import mimetypes


class UploadedFileResponse(Response):
    """
    A Response object that can be used to serve an UploadedFile instance.

    ``data`` is the ``UploadedFile`` file field value.

    ``request`` must be a Pyramid :term:`request` object.  Note
    that a request *must* be passed if the response is meant to attempt to
    use the ``wsgi.file_wrapper`` feature of the web server that you're using
    to serve your Pyramid application.

    ``cache_max_age`` is the number of seconds that should be used
    to HTTP cache this response.

    ``content_type`` is the content_type of the response.

    ``content_encoding`` is the content_encoding of the response.
    It's generally safe to leave this set to ``None`` if you're serving a
    binary file.  This argument will be ignored if you also leave
    ``content-type`` as ``None``.

    Code adapted from pyramid.response.FileResponse
    """
    def __init__(self, data, request=None, disposition='attachment',
                 cache_max_age=None, content_type=None,
                 content_encoding=None):

        if data._public_url:
            raise HTTPMovedPermanently(data._public_url)

        filename = data.filename

        content_type = content_type or getattr(data, 'content_type', None)
        if content_type is None:
            content_type, content_encoding = \
                mimetypes.guess_type(filename, strict=False)

        if content_type is None:
            content_type = 'application/octet-stream'
        # str-ifying content_type is a workaround for a bug in Python 2.7.7
        # on Windows where mimetypes.guess_type returns unicode for the
        # content_type.
        content_type = str(content_type)

        super(UploadedFileResponse, self).__init__(
            conditional_response=True,
            content_type=content_type,
            content_encoding=content_encoding
        )

        self.app_iter = FileIter(data.file, _BLOCK_SIZE)
        # assignment of content_length must come after assignment of app_iter
        self.content_length = data.file.content_length
        self.last_modified = data.file.last_modified
        if cache_max_age is not None:
            self.cache_expires = cache_max_age

        disp = '{0};filename="{1}"'.format(
            disposition, data.filename.encode('ascii', 'ignore'))
        self.headerlist.append(('Content-Disposition', disp))


@view_config(name='view', context=File, permission='view',
             renderer='kotti:templates/view/file.pt')
def view(context, request):
    return {}


@view_config(name='inline-view', context=File, permission='view')
def inline_view(context, request):
    return UploadedFileResponse(context.data, request, disposition='inline')


@view_config(name='attachment-view', context=File, permission='view')
def attachment_view(context, request):
    return UploadedFileResponse(context.data, request, disposition='attachment')


def includeme(config):
    config.scan(__name__)



import uuid
from datetime import datetime

from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy import LargeBinary
from sqlalchemy import String
from sqlalchemy import Unicode
from sqlalchemy import event
from sqlalchemy.orm import deferred

from depot.io.interfaces import FileStorage

from kotti import Base
from kotti import DBSession
from kotti.util import camel_case_to_name
from kotti.util import command

_marker = object()


class DBStoredFile(Base):
    """ :class:`depot.io.interfaces.StoredFile` implementation that stores
    file data in SQL database.

    Can be used together with :class:`kotti.filedepot.DBFileStorage` to
    implement blobs storage in the database.
    """

    __tablename__ = "blobs"

    #: Primary key column in the DB
    #: (:class:`sqlalchemy.types.Integer`)
    id = Column(Integer(), primary_key=True)
    #: Unique file id given to this blob
    #: (:class:`sqlalchemy.types.String`)
    file_id = Column(String(36), index=True)
    #: The original filename it had when it was uploaded.
    #: (:class:`sqlalchemy.types.String`)
    filename = Column(Unicode(100))
    #: MIME type of the blob
    #: (:class:`sqlalchemy.types.String`)
    content_type = Column(String(100))
    #: Size of the blob in bytes
    #: (:class:`sqlalchemy.types.Integer`)
    content_length = Column(Integer())
    #: Date / time the blob was created or last modified
    #: (:class:`sqlalchemy.types.DateTime`)
    last_modified = Column(DateTime())
    #: The binary data itself
    #: (:class:`sqlalchemy.types.LargeBinary`)
    data = deferred(Column('data', LargeBinary()))

    _cursor = 0
    _data = _marker

    def __init__(self, file_id, filename=None, content_type=None,
                 last_modified=None, content_length=None, **kwds):
        self.file_id = file_id
        self.filename = filename
        self.content_type = content_type
        self.last_modified = last_modified or datetime.now()
        self.content_length = content_length

        for k, v in kwds.items():
            setattr(self, k, v)

    def read(self, n=-1):
        """Reads ``n`` bytes from the file.

        If ``n`` is not specified or is ``-1`` the whole
        file content is read in memory and returned
        """
        if self._data is _marker:
            file_id = DBSession.merge(self).file_id
            self._data = DBSession.query(DBStoredFile.data).\
                filter_by(file_id=file_id).scalar()

        if n == -1:
            result = self._data[self._cursor:]
        else:
            result = self._data[self._cursor:self._cursor + n]

        self._cursor += len(result)

        return result

    def close(self, *args, **kwargs):
        """Implement :meth:`StoredFile.close`.
        :class:`DBStoredFile` never closes.
        """
        return

    def closed(self):
        """Implement :meth:`StoredFile.closed`.
        """
        return False

    def writable(self):
        """Implement :meth:`StoredFile.writable`.
        """
        return False

    def seekable(self):
        """Implement :meth:`StoredFile.seekable`.
        """
        return True

    def seek(self, n):
        """ Move the file cursor to position `n`

        :param n: Position for the cursor
        :type n: int
        """
        self._cursor = n

    def tell(self):
        """ Returns current position of file cursor

        :result: Current file cursor position.
        :rtype: int
        """
        return self._cursor

    @property
    def name(self):
        """Implement :meth:`StoredFile.name`.

        :result: the filename of the saved file
        :rtype: string
        """
        return self.filename

    @property
    def public_url(self):
        """ Integration with :class:`depot.middleware.DepotMiddleware`

        When supported by the storage this will provide the
        public url to which the file content can be accessed.
        In case this returns ``None`` it means that the file can
        only be served by the :class:`depot.middleware.DepotMiddleware` itself.
        """
        return None

    @classmethod
    def __declare_last__(cls):
        """ Executed by SQLAlchemy as part of mapper configuration

        When the data changes, we want to reset the cursor position of target
        instance, to allow proper streaming of data.
        """
        event.listen(DBStoredFile.data, 'set', handle_change_data)


def handle_change_data(target, value, oldvalue, initiator):
    target._cursor = 0
    target._data = _marker


def set_metadata(event):
    """Set DBStoredFile metadata based on data

    :param event: event that trigerred this handler.
    :type event: :class:`ObjectInsert` or :class:`ObjectUpdate`
    """
    obj = event.object
    obj.content_length = obj.data and len(obj.data) or 0
    obj.last_modified = datetime.now()


class DBFileStorage(FileStorage):
    """Implementation of :class:`depot.io.interfaces.FileStorage`,

    Uses `kotti.filedepot.DBStoredFile` to store blob data in an SQL database.
    """

    def get(self, file_id):
        """Returns the file given by the file_id

        :param file_id: the unique id associated to the file
        :type file_id: string
        :result: a :class:`kotti.filedepot.DBStoredFile` instance
        :rtype: :class:`kotti.filedepot.DBStoredFile`
        """

        f = DBSession.query(DBStoredFile).filter_by(file_id=file_id).first()
        if f is None:
            raise IOError
        return f

    def create(self, content, filename=None, content_type=None):
        """Saves a new file and returns the file id

        :param content: can either be ``bytes``, another ``file object``
                        or a :class:`cgi.FieldStorage`. When ``filename`` and
                        ``content_type``  parameters are not provided they are
                        deducted from the content itself.

        :param filename: filename for this file
        :type filename: string

        :param content_type: Mimetype of this file
        :type content_type: string

        :return: the unique ``file_id`` associated to this file
        :rtype: string
        """
        new_file_id = str(uuid.uuid1())
        content, filename, content_type = self.fileinfo(
            content, filename, content_type)
        if hasattr(content, 'read'):
            content = content.read()

        fstore = DBStoredFile(data=content,
                              file_id=new_file_id,
                              filename=filename,
                              content_type=content_type,
                              )
        DBSession.add(fstore)
        return new_file_id

    def replace(self, file_or_id, content, filename=None, content_type=None):
        """Replaces an existing file, an ``IOError`` is raised if the file
        didn't already exist.

        Given a :class:`StoredFile` or its ID it will replace the current
        content with the provided ``content`` value. If ``filename`` and
        ``content_type`` are provided or can be deducted by the ``content``
        itself they will also replace the previous values, otherwise the current
        values are kept.

        :param file_or_id: can be either ``DBStoredFile`` or a ``file_id``

        :param content: can either be ``bytes``, another ``file object``
                        or a :class:`cgi.FieldStorage`. When ``filename`` and
                        ``content_type`` parameters are not provided they are
                        deducted from the content itself.

        :param filename: filename for this file
        :type filename: string

        :param content_type: Mimetype of this file
        :type content_type: string
        """

        file_id = self._get_file_id(file_or_id)

        content, filename, content_type = self.fileinfo(
            content, filename, content_type)

        fstore = self.get(file_id)

        if filename is not None:
            fstore.filename = filename
        if content_type is not None:
            fstore.content_type = content_type

        if hasattr(content, 'read'):
            content = content.read()

        fstore.data = content

    def delete(self, file_or_id):
        """Deletes a file. If the file didn't exist it will just do nothing.

        :param file_or_id: can be either ``DBStoredFile`` or a ``file_id``
        """

        file_id = self._get_file_id(file_or_id)

        DBSession.query(DBStoredFile).filter_by(file_id=file_id).delete()

    def exists(self, file_or_id):
        """Returns if a file or its ID still exist.

        :return: Returns if a file or its ID still exist.
        :rtype: bool
        """

        file_id = self._get_file_id(file_or_id)

        return bool(
            DBSession.query(DBStoredFile).filter_by(file_id=file_id).count())

    def _get_file_id(self, file_or_id):
        if hasattr(file_or_id, 'file_id'):
            return file_or_id.file_id
        return file_or_id


def configure_filedepot(settings):
    from kotti.util import extract_depot_settings
    from depot.manager import DepotManager

    config = extract_depot_settings('kotti.depot.', settings)
    for conf in config:
        name = conf.pop('name')
        if name not in DepotManager._depots:
            DepotManager.configure(name, conf, prefix='')


def migrate_storage(from_storage, to_storage):
    from depot.fields.sqlalchemy import _SQLAMutationTracker
    from depot.manager import DepotManager
    from kotti.util import _to_fieldstorage
    import logging

    log = logging.getLogger(__name__)

    old_default = DepotManager._default_depot
    DepotManager._default_depot = to_storage

    for klass, props in _SQLAMutationTracker.mapped_entities.items():
        log.info("Migrating %r", klass)

        mapper = klass._sa_class_manager.mapper

        # use type column to avoid polymorphism issues, getting the same
        # Node item multiple times.
        type_ = camel_case_to_name(klass.__name__)
        for instance in DBSession.query(klass).filter_by(type=type_):
            for prop in props:
                uf = getattr(instance, prop)
                if not uf:
                    continue
                pk = mapper.primary_key_from_instance(instance)
                log.info("Migrating %s for %r with pk %r", prop, klass, pk)

                filename = uf['filename']
                content_type = uf['content_type']
                data = _to_fieldstorage(fp=uf.file,
                                        filename=filename,
                                        mimetype=content_type,
                                        size=uf.file.content_length)
                setattr(instance, prop, data)

    DepotManager._default_depot = old_default


def migrate_storages_command():  # pragma: no cover
    __doc__ = """ Migrate blobs between two configured filedepot storages

    Usage:
      kotti-migrate-storage <config_uri> --from-storage <name> --to-storage <name>

    Options:
      -h --help                 Show this screen.
      --from-storage <name>   The storage name that has blob data to migrate
      --to-storage <name>     The storage name where we want to put the blobs
    """
    return command(
        lambda args: migrate_storage(
            from_storage=args['--from-storage'],
            to_storage=args['--to-storage'],
        ),
        __doc__,
    )


def adjust_for_engine(conn, branch):
    # adjust for engine type

    if conn.engine.dialect.name == 'mysql':  # pragma: no cover
        from sqlalchemy.dialects.mysql.base import LONGBLOB
        DBStoredFile.__table__.c.data.type = LONGBLOB()

    # sqlite's Unicode columns return a buffer which can't be encoded by
    # a json encoder. We have to convert to a unicode string so that the value
    # can be saved corectly by
    # :class:`depot.fields.sqlalchemy.upload.UploadedFile`

    def patched_processed_result_value(self, value, dialect):
        if not value:
            return None
        return self._upload_type.decode(unicode(value))

    if conn.engine.dialect.name == 'sqlite':  # pragma: no cover
        from depot.fields.sqlalchemy import UploadedFileField
        UploadedFileField.process_result_value = patched_processed_result_value


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    from kotti.events import objectevent_listeners
    from kotti.events import ObjectInsert
    from kotti.events import ObjectUpdate
    from depot.fields.sqlalchemy import _SQLAMutationTracker

    from sqlalchemy.event import listen
    from sqlalchemy.engine import Engine

    listen(Engine, 'engine_connect', adjust_for_engine)

    configure_filedepot(config.get_settings())

    # Update file metadata on change of blob data
    objectevent_listeners[
        (ObjectInsert, DBStoredFile)].append(set_metadata)
    objectevent_listeners[
        (ObjectUpdate, DBStoredFile)].append(set_metadata)

    # depot's _SQLAMutationTracker._session_committed is executed on
    # after_commit, that's too late for DBFileStorage to interact with the
    # session
    event.listen(DBSession,
                 'before_commit',
                 _SQLAMutationTracker._session_committed)
